"""Cannon - hitting a target with a cannon.

Copyright (c) 2014 Grant Jenks
http://www.grantjenks.com/

Exercises
0. Explore the effect of drag.
1. Make the target smaller.
2. Make the ball larger.
3. Move the target off the ground.

"""

import sys, pygame
from pygame.locals import *
from random import randrange
from itertools import count

pygame.init()

font = pygame.font.Font(None, 14)
clock = pygame.time.Clock()
screen = pygame.display.set_mode((480, 480))

# Define colors.

white, black, purple = (255, 255, 255), (0, 0, 0), (200, 0, 200)

def reset():
    """Reset game settings."""
    global score, target
    score, target = 0, randrange(200, 450)

def redraw():
    """Redraw screen and settings."""
    global ball, xspeed, yspeed
    ball, xspeed, yspeed = None, 3.0, -6.0

    screen.fill(black)
    pygame.draw.line(screen, white, (0, 475), (480, 475), 11)
    pygame.draw.rect(screen, white, (20, 460, 10, 10))
    pygame.draw.line(screen, white, (25, 465), (35, 455), 3)
    pygame.draw.rect(screen, purple, (target, 450, 20, 20))

# Initialize start of game play.

drag = False
reset()
redraw()

for counter in count():
    if counter % 5 == 0:
        clock.tick(24)

    event = pygame.event.poll()

    # Handle pygame events.

    if event.type == pygame.QUIT:
        pygame.quit()
        sys.exit()
    elif event.type == KEYDOWN:

        # Handle user input.

        if event.key == K_SPACE:
            ball = (30.0, 460.0, xspeed, yspeed)
        elif event.key == K_UP:
            yspeed -= 0.1
        elif event.key == K_DOWN:
            yspeed += 0.1
        elif event.key == K_RIGHT:
            xspeed += 0.1
        elif event.key == K_LEFT:
            xspeed -= 0.1
        elif event.key == K_d:
            drag = (not drag)
        elif event.key == K_r:
            redraw();
        elif event.key == K_q:
            pygame.event.post(pygame.event.Event(QUIT))

    if ball is not None:

        # Update the ball's position.

        east, south, right, down = ball

        if drag:
            # Calculate the effect of air resistance (sort of).

            velocity = (right ** 2 + down ** 2) ** 0.5
            dragf = 0.005 * velocity ** 2
            down += (-down / velocity) * dragf
            right -= (right / velocity) * dragf

        # Update the velocity down due to gravity.

        down += 0.1   # gravity

        # Calculate the new position of the ball.

        east += right # xf = v * t + xi
        south += down # south_final = down_pixels/frame * 1_frame + south_initial

        ball = (east, south, right, down)

        # Draw the ball.

        ballrect = pygame.Rect(ball[0], ball[1], 3, 3)
        pygame.draw.rect(screen, white, ballrect)

        # Detect ball out of screen boundary.

        if not (0 < south < 480 and 0 < east < 480):
            ball = None

        # Detect ball hitting target.

        targetrect = pygame.Rect(target, 450, 20, 20)

        if targetrect.colliderect(ballrect):
            score += 1
            target = randrange(200, 450)
            redraw()

    # Draw text to the screen.

    pygame.draw.rect(screen, black, (0, 0, 120, 30))
    scoredisplay = font.render('score: ' + str(score), True, white)
    xdisplay = font.render('xspeed: ' + str(xspeed), True, white)
    ydisplay = font.render('yspeed: ' + str(-yspeed), True, white)
    screen.blit(scoredisplay, (0, 0))
    screen.blit(xdisplay, (0, 10))
    screen.blit(ydisplay, (0, 20))

    pygame.display.flip()



"""
Paint, for drawing shapes.

Copyright (c) 2014 Grant Jenks
http://www.grantjenks.com/

Exercises
1. Add a color.
2. Complete draw_circle.
3. Complete draw_rectangle.
4. Complete draw_triangle.
5. Add width parameter.
"""

import sys, pygame
from pygame.locals import *

pygame.init()

screen = pygame.display.set_mode((480, 480))
clock = pygame.time.Clock()

black, white = (0, 0, 0), (255, 255, 255)
red, green, blue = (255, 0, 0), (0, 255, 0), (0, 0, 255)

def draw_line(color, start, end):
    pygame.draw.line(screen, color, start, end)

def draw_square(color, start, end):
    width = end[0] - start[0]
    pygame.draw.rect(screen, color, (start[0], start[1], width, width))

def draw_circle(color, start, end):
    # TODO
    pass

def draw_rectangle(color, start, end):
    # TODO
    pass

def draw_triangle(color, start, end):
    # TODO
    pass

shapes = [[draw_square, white, (0, 0), (480, 480)]]
shape, color, start = draw_line, black, None

def draw_shapes():
    for shape in shapes:
        func, color, start, end = shape
        func(color, start, end)

while True:
    event = pygame.event.wait()

    if event.type == pygame.QUIT:
        pygame.quit()
        sys.exit()
    elif event.type == KEYDOWN:
        if event.key == K_r:
            color = red
        elif event.key == K_g:
            color = green
        elif event.key == K_u:
            color = blue
        elif event.key == K_b:
            color = black
        elif event.key == K_w:
            color = white
        elif event.key == K_l:
            shape = draw_line
        elif event.key == K_s:
            shape = draw_square
        elif event.key == K_c:
            shape = draw_circle
        elif event.key == K_e:
            shape = draw_rectangle
        elif event.key == K_t:
            shape = draw_triangle
        elif event.key == K_a:
            shapes = [[draw_square, white, (0, 0), (480, 480)]]
        elif event.key == K_q:
            pygame.event.post(pygame.event.Event(QUIT))
    elif event.type == MOUSEBUTTONDOWN and event.button == 1:
        start = pygame.mouse.get_pos()
        shapes.append([shape, color, start, start])
    elif event.type == MOUSEBUTTONUP and event.button == 1:
        start = None

    if start is not None:
        shapes[-1][-1] = pygame.mouse.get_pos()
        
    draw_shapes()
    pygame.display.flip()



"""Flappy, game inspired by Flappy Bird.

Exercises

1. Keep score.
2. Vary the speed.
3. Vary the size of the balls.
4. Allow the bird to move forward and back.

"""

from random import *
from turtle import *
from freegames import vector

bird = vector(0, 0)
balls = []

def tap(x, y):
    up = vector(0, 30)
    bird.move(up)

def inside(point):
    return -200 < point.x < 200 and -200 < point.y < 200

def draw():
    bird.y -= 5

    for ball in balls:
        ball.x -= 3

    if randrange(15) == 0:
        y = randrange(-199, 199)
        ball = vector(199, y)
        balls.append(ball)
        
    alive = inside(bird)

    for ball in balls:
        dodge = abs(ball - bird) > 15
        alive = alive and dodge

    if len(balls) > 30:
        balls.pop(0)

    clear()

    for ball in balls:
        goto(ball.x, ball.y)
        dot(20, 'black')

    goto(bird.x, bird.y)

    if alive:
        dot(10, 'green')
    else:
        dot(10, 'red')

    update()

    if alive:
        ontimer(draw, 50)

def start(x, y):
    draw()
    onscreenclick(tap)

setup(420, 420, 370, 0)
hideturtle()
up()
tracer(False)
onscreenclick(start)
done()



"""Guess a number within a range.

Exercises

1. Change the range to be from 0 to 1,000,000.
2. Can you still guess the number?
3. Print the number of guesses made.
4. Limit the number of guesses to the minimum required.

"""

import random

start = 0
end = 100
value = random.randint(start, end)

print("I'm thinking of a number between", start, 'and', end)

guess = None

while guess != value:

    try:
        guess = int(input('Guess the number: '))
    except:
        print('Whoops! Be sure the number contains only digits.')
        continue

    if guess < value:
        print('Higher.')
    elif guess > value:
        print('Lower.')

print('Congratulations! You guessed the right answer:', value)



"""
Maze game.

Copyright (c) 2014 Grant Jenks
http://www.grantjenks.com/

Exercises
1. Change where the end is placed.
2. Change where the player starts.
3. Display only the player's immediate surroundings (no history).
"""

import sys, pygame
import random
from pygame.locals import *

pygame.init()

tiles, block = 24, 20
size = width, height = (tiles * block), (tiles * block + 10)

dirs = up, right, down, left = (-tiles, 1, tiles, -1)
view = (up, right, down, left, up - 1, up + 1, down - 1, down + 1)

hall, wall, boundary, extra = 0, 1, 2, 3

red, green, blue = (255, 0, 0), (0, 255, 0), (0, 0, 255)
white, grey, black = (255, 255, 255), (100, 100, 100), (0, 0, 0)

font = pygame.font.Font(None, 14)
clock = pygame.time.Clock()
screen = pygame.display.set_mode(size)

def make_maze():
    global maze, player, end, moves

    # Create the initial, empty maze.

    maze = [boundary] * tiles
    maze += ([boundary] + [wall] * (tiles - 2) + [boundary]) * (tiles - 2)
    maze += [boundary] * tiles

    # Make the maze. This is a randomized version of Prim's algorithm
    # to build a minimum spanning tree.

    player = tiles + 1
    maze[player] = 0
    frontier = [tiles + 2, 2 * tiles + 1]
    while len(frontier) > 0:
        pos = random.randrange(len(frontier))
        frontier[pos], frontier[-1] = frontier[-1], frontier[pos]
        spot = frontier.pop()

        if maze[spot] != wall: continue

        if map(lambda diff: maze[spot + diff], dirs).count(0) != 1:
            continue

        maze[spot] = hall

        frontier.extend(map(lambda diff: spot + diff, dirs))

    # End goal should be farthest reachable hall.

    frontier = [player]
    while True:
        next = []
        for front in frontier:
            maze[front] = extra
            next.extend(filter(lambda pos: maze[pos] == hall,
                               map(lambda diff: front + diff, dirs)))
        if next:
            frontier = next
        else:
            last = random.choice(frontier)
            break

    for pos, val in enumerate(maze):
        if val == extra: maze[pos] = hall

    moves = 0
    end = last

def draw_tile(func, color, pos):
    left = (pos % tiles) * block
    top = (pos / tiles) * block
    func(screen, color, (left, top, block, block))

def draw_maze():
    for pos, val in enumerate(maze):
        color = white if val == hall else black
        draw_tile(pygame.draw.rect, color, pos)

def draw_view():
    for diff in view:
        pos = player + diff
        color = white if maze[pos] == hall else black
        draw_tile(pygame.draw.rect, color, pos)

    draw_tile(pygame.draw.rect, white, player)
    draw_tile(pygame.draw.rect, green, end)
    draw_tile(pygame.draw.ellipse, blue, player)

    pygame.draw.rect(screen, (0, 0, 0), (0, width, width, 10))
    surface = font.render(str(moves), True, (255, 255, 255))
    screen.blit(surface, (5, tiles * block))

    pygame.display.flip()

def restart():
    make_maze()
    pygame.draw.rect(screen, grey, (0, 0, width, height))
    draw_view()

restart()

while True:
    clock.tick(12)

    event = pygame.event.poll()
    move_dir = None

    if event.type == pygame.QUIT:
        pygame.quit()
        sys.exit()
    elif event.type == KEYDOWN:
        if event.key == K_UP:
            move_dir = up
        elif event.key == K_RIGHT:
            move_dir = right
        elif event.key == K_DOWN:
            move_dir = down
        elif event.key == K_LEFT:
            move_dir = left
        elif event.key == K_r:
            restart()
        elif event.key == K_s:
            draw_maze()
            draw_view()
        elif event.key == K_q:
            pygame.event.post(pygame.event.Event(QUIT))

    if player == end: continue

    if move_dir is not None:
        if maze[player + move_dir] == hall:
            player += move_dir
            moves += 1
        draw_view()

    pygame.display.flip()



"""
Cups - a game of tracking a cup.

Copyright (c) 2014 Grant Jenks
http://www.grantjenks.com/

Exercises
1. Increase the number of cups.
2. Increase the frame rate.
3. Increase the cup rotation speed.
4. Increase the number of swaps.
5. Change the circles to squares.
6. How could you allow multiple pairs of cups to rotate at the same time?
"""

import sys, pygame, random
from pygame.locals import *
from math import pi, sin, cos

pygame.init()

cup_count, cup_width, margin, fps = 3, 80, 20, 20
block = cup_width + margin
width = block * cup_count

clock = pygame.time.Clock()
screen = pygame.display.set_mode((width, width))

white, blue, red = (255, 255, 255), (0, 0, 255), (255, 0, 0)

def draw_cup(index, color=blue, size=2):
    xpos = int(index * block + block / 2)
    ypos = int(width / 2)
    pygame.draw.circle(screen, color, (xpos, ypos), cup_width / size)

def draw_cups():
    pygame.draw.rect(screen, white, (0, 0, width, width))
    for index in range(cup_count):
        draw_cup(index)

def animate_start():
    for repeat in range(fps):
        draw_cups()
        for idx, val in enumerate(cups):
            if val == cup:
                draw_cup(idx, red, 4)
        yield

def animate_swap(left, right):
    if left > right:
        left, right = right, left

    direction = random.choice([1, -1])
    radius = (right - left) * (block / 2)
    center = (left + right) / 2.0 * block + (block / 2)

    for angle in range(0, 180, 10):
        angle *= direction
        pygame.draw.rect(screen, white, (0, 0, width, width))
        for index in range(cup_count):
            if index == left or index == right:
                offset = 0 if index == right else 180
                radians = (angle + offset) * pi / 180
                xpos = int(center + radius * cos(radians))
                ypos = int(width / 2 - radius * sin(radians))
                pygame.draw.circle(screen, blue, (xpos, ypos), cup_width / 2)
            else:
                draw_cup(index)
        yield

    cups[left], cups[right] = cups[right], cups[left]

def reset():
    global cups, cup, animation, swaps
    cups = list(range(cup_count))
    cup = random.randrange(cup_count)
    animation = animate_start()
    swaps = random.randrange(10, 20)

reset()

while True:
    clock.tick(fps)

    event = pygame.event.poll()

    if event.type == pygame.QUIT:
        pygame.quit()
        sys.exit()
    elif event.type == KEYDOWN:
        if event.key == K_r:
            reset()
        elif event.key == K_q:
            pygame.event.post(pygame.event.Event(QUIT))

    if animation is None:
        if swaps > 0:
            start = random.randrange(cup_count - 1)
            other = random.randrange(1, cup_count - start)
            animation = animate_swap(start, start + other)
        elif swaps > -(fps * 5):
            draw_cups()
        else:
            animation = animate_start()
        swaps -= 1
    else:
        try:
            animation.next()
        except:
            animation = None

    pygame.display.flip()
